/*
 * sxkey.c, part of gSXprog
 * The functions in this module are common between old and new keys.
 *
 * Copyright 2001-2003 Eric Smith <eric@brouhaha.com>
 *
 * $Id: sxkey.c,v 1.5 2003/11/18 13:59:42 garlick Exp $
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include "gsxprog.h"
#include "sxkey.h"

/*
 * NOTE on crypto stuff:
 * encrypt()/decrypt() are called by sx_send_byte()/sx_get_byte().
 * On old keys without crypto, sx_setkeys() is never called and 
 * encrypt()/decrypt() are identity functions.
 */

static int _crypto_enabled = false;

/* values used in stream encryption */
static u8 _crypto_key_1 = 0x00;
static u8 _crypto_key_2 = 0x00;


void sx_setkeys(u8 c1, u8 c2)
{
    _crypto_key_1 = c1 ^ 0x37;
    _crypto_key_2 = c2 ^ 0x95;
    _crypto_enabled = true;
}

void sx_unsetkeys(void)
{
    _crypto_enabled = false;
}

static void _munge_keys(void)
{
    u8 temp;

    temp = _crypto_key_1 ^ _crypto_key_2;
    temp ^= (temp << 6);

    temp = (_crypto_key_1 >> 1) | (temp & 0x80);

    _crypto_key_2 = (_crypto_key_1 & 0x01) | (_crypto_key_2 << 1);

    _crypto_key_1 = temp;
}

static u8 _decrypt(u8 c)
{
    u8 p;

    if (!_crypto_enabled)
        return (c);

    p = _crypto_key_1 - (c ^ _crypto_key_2);
    _munge_keys();
    return (p);
}

static u8 _encrypt(u8 p)
{
    u8 c;

    if (!_crypto_enabled)
        return (p);

    _munge_keys();
    c = _crypto_key_1 ^ (p - _crypto_key_2);
    return (c);
}

boolean sx_serial_open(char *device_name, speed_t bit_rate, int *fd)
{
    int r;
    struct termios t;

    *fd = open(device_name, O_RDWR | O_NOCTTY | O_NDELAY);
    if (*fd < 0) {
        fprintf(stderr,"Unable to open device!\n");
        perror(device_name);
        return false;
    }

    r = tcgetattr(*fd, &t);
    if (r < 0)
        return false;


    cfmakeraw(&t);

    r = cfsetospeed(&t, bit_rate);
    if (r < 0)
        return false;

    r = cfsetispeed(&t, 0);
    if (r < 0)
        return false;

	t.c_cflag = CLOCAL | CREAD | CS8;
	t.c_iflag = IGNPAR | IGNBRK;
	t.c_oflag = 0;
    r = tcsetattr(*fd, TCSANOW, &t);
    if (r < 0)
        return false;

    fprintf(stderr,"Device Opened!\n");
    
    return true;
}

void sx_serial_close(int fd)
{
    close(fd);
}

boolean sx_serial_flush(int fd)
{
    if (tcflush(fd, TCIOFLUSH) < 0) {
        perror("tcflush");
        return false;
    }
    return true;
}

boolean sx_serial_read(int fd, int ms_timeout, u8 * c)
{
    fd_set rfds;
    struct timeval tv;
    ssize_t len;
    int retval;

    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    tv.tv_sec = ms_timeout / 1000;
    tv.tv_usec = (ms_timeout % 1000) * 1000;

    retval = select(fd + 1, &rfds, NULL, NULL, &tv);

    if (retval < 0) {
        perror("select");
        return false;
    }

    /* $$$ why isn't select() working? */
    if (!FD_ISSET(fd, &rfds))
        return false;

    len = read(fd, c, 1);

    if (len != 1) {
        perror("read");
        return false;
    }

    return true;
}

boolean sx_serial_write(int fd, u8 c)
{
    ssize_t len;

    len = write(fd, &c, 1);
    if (len != 1) {
        perror("write");
        return false;
    }
    return true;
}

void sx_ms_delay(int ms)
{
    struct timeval tv;
    int retval;

    tv.tv_sec = 0;
    tv.tv_usec = ms * 1000;

    retval = select(0, NULL, NULL, NULL, &tv);
    if (retval < 0)
        perror("select");
}

/* send a character to SX-Key and make sure it is echoed */
boolean sx_send_byte(int fd, u8 c)
{
    boolean r;
    u8 read_char;

    c = _encrypt(c);

    r = sx_serial_write(fd, c);
    if (!r)
        return (r);

    r = sx_serial_read(fd, SXKEY_ECHO_TIMEOUT, &read_char);
    if (!r)
        return (r);

    if (read_char != c)
        return false;

    return true;
}

boolean sx_send_word(int fd, u16 w)
{
    int r;

    r = sx_send_byte(fd, w & 0xff);
    if (r)
        r = sx_send_byte(fd, (w >> 8) & 0xff);
    return (r);
}

boolean sx_send_str(int fd, char *s)
{
    int r = 1;

    while (r && *s)
        r &= sx_send_byte(fd, *s++);
    return (r);
}

/* get a character from SX-Key */
boolean sx_get_byte(int fd, int ms_timeout, u8 * c)
{
    boolean r;
    u8 tmpc;

    r = sx_serial_read(fd, ms_timeout, &tmpc);
    if (!r)
        return (r);

    tmpc = _decrypt(tmpc);
    if (c)
        (*c) = tmpc;

    return true;
}

boolean sx_get_word(int fd, int ms_timeout, u16 * w)
{
    int r;
    u8 cl, ch;

    r = sx_get_byte(fd, ms_timeout, &cl);
    if (!r)
        return false;
    r = sx_get_byte(fd, ms_timeout, &ch);
    if (!r)
        return false;
    if (w)
        *w = (ch << 8) + cl;
    return (r);
}

boolean sx_expect_str(int fd, int ms_timeout, char *s)
{
    int r = false;
    u8 c;

    while (*s) {
        r = sx_get_byte(fd, ms_timeout, &c);
        if (!r)
            return (0);
        if (c != (u8) * s)
            return (0);
        s++;
    }
    return (r);
}

/*
 * Determine size of ROM area (in words) by looking at DEVICE/FUSEX values.
 */
int sx_rom_size(sx_t * sx)
{
    int sizes_sx28[] = { 0x200, 0x400, 0x800, 0x800 };
    int size = 0x1000;          /* SX 48/52 */

    if (SX_28_MODE(sx->device))
        size = sizes_sx28[sx->rom[FUSEX_LOC] & 2];

    return size;
}

static void _sx28_show_fuse(sx_t * sx)
{
    char *f_turbo[] = { "TURBO ", "" };
    char *f_sync[] = { "SYNC ", "" };
    char *f_irc[] = { "IRC ", "" };
    char *f_div[] = { "OSC4MHZ ", "OSC1MHZ ", "OSC128KHZ ", "OSC32KHZ " };
    char *f_ifbd[] = { "", "IFBD " };
    char *f_cp[] = { "PROTECT ", "" };
    char *f_wdte[] = { "", "WATCHDOG " };
    char *f_osc[] = { "OSCLP1 ", "OSCLP2 ", "OSCXT1 ", "OSCXT2 ",
        "OSCHS1 ", "OSCHS2 ", "OSCHS3 ", "OSCRC "
    };

    char *fx_pins[] = { "SX18/20 ", "SX28 " };
    char *fx_optionx[] = { "STACKX_OPTIONX ", "" };
    char *fx_cf[] = { "CARRYX ", "" };
    char *fx_bor[] = { "BOR42 ", "BOR26 ", "BOR22 ", "" };
    char *fx_bp[] = { "1PAGE/1BANK ", "2PAGE/2BANK ", "4PAGE/4BANK ",
        "4PAGE/8BANK "
    };

    u16 f = sx->rom[FUSE_LOC];
    u16 fx = sx->rom[FUSEX_LOC];
    int irc_ena = (((f >> 7) & 1) == 0);        /* multi-use fuse bits */

    printf("device: 0x%-3.3hx SX18/28\n", sx->device);

    printf("fuse:   0x%-3.3hx ", f);
    printf("%s", f_turbo[(f >> 11) & 1]);
    printf("%s", f_sync[(f >> 10) & 1]);
    printf("%s", f_irc[(f >> 7) & 1]);
    if (irc_ena)
        printf("%s", f_div[(f >> 5) & 3]);
    else
        printf("%s", f_ifbd[(f >> 6) & 1]);
    printf("%s", f_cp[(f >> 3) & 1]);
    printf("%s", f_wdte[(f >> 2) & 1]);
    if (!irc_ena)
        printf("%s", f_osc[(f & 3) | ((f >> 3) & 4)]);
    printf("\n");

    printf("fusex:  0x%-3.3hx ", fx);
    printf("IRCTRIM%d ", ((fx >> 8) & 3) | ((fx >> 11) & 1));
    printf("%s", fx_pins[(fx >> 10) & 1]);
    printf("%s", fx_optionx[(fx >> 7) & 1]);
    printf("%s", fx_cf[(fx >> 6) & 1]);
    printf("%s", fx_bor[(fx >> 4) & 3]);
    printf("BORTRIM%d ", (fx >> 2) & 3);
    printf("%s", fx_bp[fx & 3]);
    printf("\n");
}

/* XXX untested - jg 5/28/02 */
static void _sx52_show_fuse(sx_t * sx)
{
    char *f_sync[] = { "SYNC ", "" };
    char *f_irc[] = { "IRC ", "" };
    char *f_div[] = { "OSC4MHZ ", "OSC1MHZ ", "OSC128KHZ ", "OSC32KHZ " };
    char *f_ifbd[] = { "", "IFBD " };
    char *f_xtl[] = { "", "XTLBUF_EN " };
    char *f_cp[] = { "PROTECT ", "" };
    char *f_wdte[] = { "", "WATCHDOG " };
    char *f_osc[] = { "OSCLP1 ", "OSCLP2 ", "OSCXT1 ", "OSCXT2 ",
        "OSCHS1 ", "OSCHS2 ", "OSCHS3 ", "OSCRC "
    };

    char *fx_sleep[] = { "", "SLEEPCLOCK " };
    char *fx_cf[] = { "CARRYX ", "" };
    char *fx_bor[] = { "BOR42 ", "BOR26 ", "BOR22 ", "" };
    char *fx_drt[] = { "DRT60MS ", "DRT960MS ", "DRT60US ", "DRT18MS " };

    u16 f = sx->rom[FUSE_LOC];
    u16 fx = sx->rom[FUSEX_LOC];
    int irc_ena = (((f >> 7) & 1) == 0);        /* multi-use fuse bits */

    printf("device: 0x%-3.3hx SX48/52\n", sx->device);

    printf("fuse:   0x%-3.3hx ", f);
    printf("%s", f_sync[(f >> 10) & 1]);
    printf("%s", f_irc[(f >> 7) & 1]);
    if (irc_ena)
        printf("%s", f_div[(f >> 5) & 3]);
    else
        printf("%s", f_ifbd[(f >> 6) & 1]);
    printf("%s", f_xtl[(f >> 4) & 1]);
    printf("%s", f_cp[(f >> 3) & 1]);
    printf("%s", f_wdte[(f >> 2) & 1]);
    if (!irc_ena)
        printf("%s", f_osc[(f & 3) | ((f >> 3) & 4)]);
    printf("\n");

    printf("fusex:  0x%-3.3hx ", fx);
    printf("IRCTRIM%d ", ((fx >> 8) & 3) | ((fx >> 11) & 1));
    printf("%s", fx_sleep[(fx >> 10) & 1]);
    printf("%s", fx_cf[(fx >> 6) & 1]);
    printf("%s", fx_bor[(fx >> 4) & 3]);
    printf("BORTRIM%d ", (fx >> 2) & 3);
    printf("%s", fx_drt[fx & 3]);
    printf("\n");
}

void sx_show_fuse_bits(sx_t * sx)
{
    if (SX_28_MODE(sx->device))
        _sx28_show_fuse(sx);
    else
        _sx52_show_fuse(sx);
}

void sx_show(sx_t * sx)
{
    int i, size_rom = sx_rom_size(sx);

    sx_show_fuse_bits(sx);

    printf("rom:");
    for (i = 0; i < size_rom; i++) {
        if ((i & 7) == 0)
            printf("\n%03x:", i);
        printf(" %04x", sx->rom[i]);
    }
    printf("\n");

    printf("id:");
    for (i = 0; i < SX_ID_SIZE; i++) {
        if ((i & 7) == 0)
            printf("\n%03x:", i);
        printf(" %04x", sx->rom[ID_LOC + i]);
    }
    printf("\n");
}

boolean sx_compare(sx_t * sx1, sx_t * sx2, boolean ignoretrim, boolean idskip)
{
    boolean r = 1;
    int i;
    int rom_size = sx_rom_size(sx1);
    u16 fusex_ignore = ignoretrim ? ~(IRCTRIM0 | IRCTRIM1 | IRCTRIM2) : 0xffff;

    if ((sx1->rom[FUSEX_LOC] & fusex_ignore) != 
            (sx2->rom[FUSEX_LOC] & fusex_ignore)) {
        fprintf(stderr, "%s: FUSEX mismatch: expected %03x, got %03x\n",
                progname, sx1->rom[FUSEX_LOC], sx2->rom[FUSEX_LOC]);
        r = 0;
    }

    if (sx1->rom[FUSE_LOC] != sx2->rom[FUSE_LOC]) {
        fprintf(stderr, "%s: FUSE mismatch: expected %03x, got %03x\n",
                progname, sx1->rom[FUSE_LOC], sx2->rom[FUSE_LOC]);
        r = 0;
    }

    for (i = 0; i < rom_size; i++) {
        if (sx1->rom[i] != sx2->rom[i]) {
            fprintf(stderr,
                    "%s: ROM word %03x mismatch: expected %03x, got %03x\n",
                    progname, i, sx1->rom[i], sx2->rom[i]);
            r = 0;
            break;
        }
    }

    if(!idskip) {
        for (i = 0; i < SX_ID_SIZE; i++) {
            if (sx1->rom[ID_LOC + i] != sx2->rom[ID_LOC + i]) {
                fprintf(stderr,
                        "%s: ID word %03x mismatch: expected %03x, got %03x\n",
                        progname, i, sx1->rom[ID_LOC + i],
                        sx2->rom[ID_LOC + i]);
                r = 0;
                break;
            }
        }
    }

    return (r);
}

/*
 * em:Local Variables: 
 * em:indent-tabs-mode:nil
 * em:c-basic-offset:4
 * em:End:
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
